<div class="github-star-badge">
    <a href="https://github.com/{{ github_user }}/{{ github_repo }}">
        <img src="https://img.shields.io/github/stars/{{ github_user }}/{{ github_repo }}?style=social&label=Star%20on%20%20GitHub%E2%AD%90" alt="Support us with a star on GitHub">
    </a>
</div>
